# -*- coding: utf-8 -*-

#    Copyright 2010 Tim Prosser

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from osgb_ui import Ui_OSGB
from PyQt4 import QtCore, QtGui
from QtMobility.Location import *
from math import *
from datetime import *
import linecache

def helmert(x,y,z,tx,ty,tz,s,rx,ry,rz):
    # Do the transform
    x1 = tx + ((1.0+s) * x) + (-rz * y) + (ry * z)
    y1 = ty + (rz * x) + ((1.0+s) * y) + (-rx * z)
    z1 = tz + (-ry * x) + (rx * y) + ((1.0+s) * z)   
    return x1,y1,z1

def latlong_2_cart(lat,long,height,a,b):
    e2 = (a**2 - b**2) / a**2
    
    theta = (lat  / 180.0) * pi
    lamb = (long / 180.0) * pi

    v = a / sqrt(1.0 - e2 * (sin(theta)**2) )
    x = (v + height) * cos(theta) * cos(lamb)
    y = (v + height) * cos(theta) * sin(lamb)
    z = ((1.0 - e2) * v + height ) * sin(theta)
    return x,y,z

def cart_2_latlong(x,y,z,a,b):
    e2 = (a**2 - b**2) / a**2
    p = sqrt(x*x + y*y)
    lamb = atan(y/x)
    
    phi0 = atan( z / (p * (1.0 - e2)) )
    while 1:
        v = a / sqrt( 1.0 - e2 * (sin(phi0)**2) )
        phi = atan( (z + e2*v*sin(phi0)) / p )
        if (phi0-phi) < 0.00001:
            break
        else:
            phi0 = phi
            
    height = (p / cos(phi)) - v 

    long = lamb * 180/pi
    lat = phi * 180/pi
    return(lat,long,height)

def itrs_2_etrs89(lat,long,height):
    today = datetime.utcnow()
    refdate = datetime(1989, 1, 1, 0, 0, 0)
    deltat = (today - refdate).days / 365.2425
    
    #ITRS2000 to ETRS89 datum transformation
    (x,y,z) = latlong_2_cart(lat,long,height,6378137.0000,6356752.3141)
    tx = 0.054
    ty = 0.051
    tz = -0.048
    s = 0.0
    rx = 0.000054 * deltat * pi / (3600 * 180)
    ry = 0.000518 * deltat * pi / (3600 * 180)
    rz = -0.000781 * deltat * pi / (3600 * 180)
    (x1,y1,z1) = helmert(x,y,z,tx,ty,tz,s,rx,ry,rz)
    (lat1,long1,height1) = cart_2_latlong(x1,y1,z1,6378137.0000,6356752.3141)
    return lat1,long1,height1

def ll_2_osgben(ellipsoid, etrslat, etrslong):
    #Transform decimal latitude longitude pair to easting and northing
    #There is no datum transformation 
    phi = (etrslat / 180) * pi
    lamb = (etrslong / 180) * pi
    
    if ellipsoid == "GRS80":
        #Constants for the GRS80 (WGS84) ellipsoid
        a = 6378137.0000  #Semi-major axis
        b = 6356752.3141  #Semi-minor axis
    elif ellipsoid == "Airy1830":
        #Constants for Airy 1830 ellipsoid
        a = 6377563.3960  #Semi-major axis
        b = 6356256.9100  #Semi-minor axis
    else:
        #Oops
        return None
            
    F0 = 0.9996012717 #Scale factor
    
    phi0 = (49.0 / 180.0) * pi #True origin latitude
    lamb0 = (-2.0 / 180.0) * pi #True origin longitude
    
    E0 = 400000.0 #True origin easting
    N0 = -100000.0 #True origin northing
    
    #Eccentricity constant squared
    e2 = (a**2 - b**2) / a**2
    
    n = (a-b)/(a+b)
    nu = a * F0 * (1.0 - (e2 * (sin(phi))**2.0))**(-0.5)
    rho = a * F0 * (1.0 - e2) * ((1.0 - (e2 * (sin(phi)**2.0)))**-1.5)
    eta2 = (nu/rho) - 1.0
    M = b * F0 * ( \
                   (1.0 + n + ((5.0/4.0)*(n**2.0)) + ((5.0/4.0)*(n**3.0))) * (phi-phi0) \
                   - (3.0*n + (3.0*(n**2.0)) + ((21.0/8.0)*(n**3.0))) * sin(phi-phi0) * cos(phi+phi0) \
                   + ((15.0/8.0)*(n**2.0) + (15.0/8.0)*(n**3.0)) * sin(2.0*(phi-phi0)) * cos(2.0*(phi+phi0)) \
                   - ((35.0/24.0)*(n**3.0)) *  sin(3.0*(phi-phi0)) * cos(3.0*(phi+phi0)) \
                   )
    I = M + N0
    II = (nu / 2.0) * sin(phi) * cos(phi)
    III = (nu / 24.0) * sin(phi) * (cos(phi)**3.0) * (5.0 - (tan(phi)**2.0) + (9.0 * eta2))
    IIIA = (nu / 720.0) * sin(phi) * (cos(phi)**5.0) * (61.0 - 58.0 * (tan(phi)**2.0) + (tan(phi)**4.0))
    IV = nu * cos(phi)
    V = (nu / 6.0) * (cos(phi)**3.0) * ((nu/rho) - (tan(phi)**2.0))
    VI = (nu / 120.0) * (cos(phi)**5.0) * (5.0 - 18.0*(tan(phi)**2.0) + (tan(phi)**4.0) + 14.0*eta2 - 58.0*(tan(phi)**2.0)*eta2) 
    
    N = I + II*((lamb-lamb0)**2.0) + III*((lamb-lamb0)**4.0) + IIIA*((lamb-lamb0)**6.0)
    E = E0 + IV*(lamb-lamb0) + V*((lamb-lamb0)**3.0) + VI*((lamb-lamb0)**5.0)
    return E,N

def osgno2(x,y,h):
    #Ordnance survey OSGN02 transform from ETRS89
    fn = "OSTN02_OSGM02_GB.txt"

    x0 = floor(x/1000.0) * 1000.0
    y0 = floor(y/1000.0) * 1000.0
    east_index = int(floor(x/1000.0))
    north_index = int(floor(y/1000.0))
    
    lineno0,e0,n0,se0,sn0,sg0,dt0 = map(float, linecache.getline(fn,(east_index + north_index*701 + 1)).split(","))
    lineno1,e1,n1,se1,sn1,sg1,dt1 = map(float, linecache.getline(fn,(east_index + 1 + north_index*701 + 1)).split(","))
    lineno2,e2,n2,se2,sn2,sg2,dt2 = map(float, linecache.getline(fn,(east_index + 1 + (north_index + 1)*701 + 1)).split(","))
    lineno3,e3,n3,se3,sn3,sg3,dt3 = map(float, linecache.getline(fn,(east_index + (north_index + 1)*701 + 1)).split(","))
       
    dx = x - x0
    dy = y - y0
    
    t = dx / 1000.0
    u = dy / 1000.0
    
    se = (1.0-t)*(1.0-u)*se0 + t*(1.0-u)*se1 + t*u*se2 + (1.0-t)*u*se3
    sn = (1.0-t)*(1.0-u)*sn0 + t*(1.0-u)*sn1 + t*u*sn2 + (1.0-t)*u*sn3
    sg = (1.0-t)*(1.0-u)*sg0 + t*(1.0-u)*sg1 + t*u*sg2 + (1.0-t)*u*sg3
    
    E = x + se
    N = y + sn
    H = h - sg
    
    return E, N, H

def GPS_2_OS(lat,long,height):
    (elat,elong,eheight) = itrs_2_etrs89(lat,long,height)
    (easting,northing) = ll_2_osgben("GRS80",elat,elong)
    (oseasting,osnorthing,osheight) = osgno2(easting,northing,height)
    return oseasting,osnorthing,osheight

def EN_2_GR(E,N):
    ref100 = ["VWXYZ","QRSTU","LMNOP","FGHJK","ABCDE"]
    ref500 = ["ST","NO", "HJ"]
    
    #Get the 500km and 100km coordinates and pull out the relevant character from the array for each
    intN100 = int(((N % 500000) - (N % 100000)) / 100000)
    intN500 = int((N - intN100*100000 - (N % 100000)) / 500000)
    intE100 = int(((E % 500000) - (E % 100000)) / 100000)
    intE500 = int((E - intE100*100000 - (E % 100000)) / 500000)
    
    s500 = ref500[intN500][intE500]
    s100 = ref100[intN100][intE100]
    
    sE = int(round(E)) % 100000
    sN = int(round(N)) % 100000

    return s500 + s100 + " %05d %05d" % (sE,sN)

class KML:
    def __init__(self, filepath):
        self.filepath = filepath
        file = open(filepath,"w")
        file.write(
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"\
        "<kml xmlns=\"http://www.opengis.net/kml/2.2\">\n"\
        "<Document>\n"\
        "<name>OSGB Path</name>\n"\
        "<Style id='redLine'>\n"\
        "  <LineStyle>\n"\
        "    <color>ff0000ff</color>\n"\
        "    <width>2</width>\n"\
        "  </LineStyle>\n"\
        "</Style>\n")
        file.close()
        self.pathcount = 0
        self.record_path = False

    def start_path(self):
        if self.record_path == False:
            self.pathcount = self.pathcount + 1
            self.gpspath = []
            self.record_path = True

    def add_to_path(self, latitude, longitude, altitude, ztime):
        if self.record_path:
            self.gpspath.append([longitude, latitude, altitude, ztime])

    def end_path(self):
        if self.record_path and len(self.gpspath) > 0:
            file = open(self.filepath,"a")
            file.write(
            "<Placemark>\n"\
            "  <description>OSGB Path No %d</description>\n" % self.pathcount + \
            "  <name>%d</name>\n" % self.pathcount +\
            "  <styleUrl>redLine</styleUrl>\n"\
            "  <LineString id=\"%d\"" % self.pathcount + ">\n"\
            "  <altitudeMode>clampToGround</altitudeMode>\n"\
            "  <coordinates>\n")
            for coord in self.gpspath:
                file.write("    %f,%f,%f\n" % (coord[0], coord[1],coord[2]))
            file.write(
            "  </coordinates>\n"\
            "  </LineString>\n"\
            "</Placemark>\n")
            file.close()
            self.record_path = False

    def close(self):
        if self.record_path:
            self.end_path()
        file = open(self.filepath,"a")
        file.write(
        "</Document>\n"\
        "</kml>")
        file.close()

class SetCoordinates:
    def __init__(self,MainWindow):
        self.source = QGeoPositionInfoSource.createDefaultSource(None)
        if self.source is not None:
            self.source.positionUpdated.connect(self.updatePosition)
            self.source.startUpdates()
        self.satellites = QGeoSatelliteInfoSource.createDefaultSource(None)
        if self.satellites is not None:
            self.satellites.satellitesInViewUpdated.connect(self.satellitesViewUpdate)
            self.satellites.satellitesInUseUpdated.connect(self.satellitesUseUpdate)
            self.satellites.startUpdates()
            
        self.lastsatupdate={}
        self.vissat={}
        self.actsat={}
        self.isfix=False
			
    def updatePosition(self, update=None):
        if update:
            latitiude = update.coordinate().latitude()
            longitude = update.coordinate().longitude()
            ALT = update.coordinate().altitude()
            GR_Error = update.attribute(QGeoPositionInfo.HorizontalAccuracy)
            ALT_Error = update.attribute(QGeoPositionInfo.VerticalAccuracy)
            GSpeed = update.attribute(QGeoPositionInfo.GroundSpeed)
            ztime = update.timestamp()
            (x,y,z) = GPS_2_OS(latitiude,longitude,ALT)
            GR = EN_2_GR(x,y)
            MainWindow.setGR(GR,GR_Error,ALT,ALT_Error,ztime,latitiude,longitude)
            MainWindow.setSpeed(GSpeed)
            if not self.isfix:
                self.isfix = True
                
    def satellitesUseUpdate(self,update=None):
        if update:
            self.actsat = {}
            for sat in update:
                ss = max([0.0,sat.signalStrength()])
                ss = min([ss,40.0])
                ss = (ss / 40.0) * 100.0
                self.actsat[sat.prnNumber()] = ss
                            
    def satellitesViewUpdate(self,update=None):
        if update:
            #if len(self.lastsatupdate) <= len(update):
            self.vissat = {}
            for sat in update:
                ss = max([0.0,sat.signalStrength()])
                ss = min([ss,40.0])
                ss = (ss / 40.0) * 100.0
                self.vissat[sat.prnNumber()] = ss
            #else:
            #    self.actsat = {}
            #    for sat in update:
            #        self.actsat[sat.prnNumber()] = sat.signalStrength()
            MainWindow.updateSat(self.vissat,self.actsat)
            self.lastsatupdate = update
              
class OSGBMainWindow(QtGui.QMainWindow):
    def __init__(self, parent=None):
        QtGui.QMainWindow.__init__(self, parent)
        self.ui = Ui_OSGB()
        self.ui.setupUi(self)
        self.vissat = {}
        self.actsat = []
        self.KML = KML("osgb.kml")
        self.KML.start_path()
        self.connect(self, QtCore.SIGNAL('triggered()'),self.closeEvent)

    def closeEvent(self, event):
        self.KML.close()
        self.deleteLater()

    def setGR(self, GR, GR_Error, ALT, ALT_Error, ztime, latitude, longitude):
        self.ui.GR.setText(GR + "<font size='-16' color=#555> %dm</font>" % GR_Error)
        self.ui.ALT.setText("%dm" % ALT + " <font size='-16' color=#555> %dm</font>" % ALT_Error)
        self.ui.TIME.setText(ztime.toString())
        self.KML.add_to_path(latitude,longitude,ALT,ztime)

    def setSpeed(self,GSpeed):
        #Convert to kmh
        GSpeed = GSpeed * 3600 / 1000
        self.ui.Speed.setText("%.1f km/h" % GSpeed)
    
    def updateSat(self,vissat,actsat):
        for i, key in enumerate(sorted(vissat.keys())):
            pbar = self.ui.satframe.findChild(QtGui.QProgressBar,"SatStr_" + str(i+1))
            plbl = self.ui.satframe_lbl.findChild(QtGui.QLabel,"label_" + str(i+1))
            if key in actsat.keys():
                pbar.setStyleSheet('QProgressBar {background-color:#111;color:#111;border: 8px solid #111;} QProgressBar::chunk{background-color: #008800;}')
            else:
                pbar.setStyleSheet('QProgressBar {background-color:#111;color:#111;border: 8px solid #111;} QProgressBar::chunk{background-color: #666666}')
            pbar.setProperty("value", vissat[key])
            plbl.setText("%d" % key)    
        for j in range(len(vissat.keys()),12):
            self.ui.satframe.findChild(QtGui.QProgressBar,"SatStr_" + str(j+1)).setProperty("value", 0)
            self.ui.satframe_lbl.findChild(QtGui.QLabel,"label_" + str(j+1)).setText("")

if __name__ == "__main__":
  app = QtGui.QApplication(sys.argv)
  MainWindow = OSGBMainWindow()
  MainWindow.show()
  locator = SetCoordinates(MainWindow)
  sys.exit(app.exec_())
